/* prs_mp.c (multiprecision and rational arithmetic)
   Ambros Marzetta
   derived from David Avis' lrs.c, version 2.5i, 1995-08-15
  
 * $Source: /home/schilte/marzetta/work/pd/RCS/prs_mp.c,v $
 * $Revision: 1.3 $
 * $Date: 1997/04/18 14:58:00 $
 * $Log: prs_mp.c,v $
 * Revision 1.3  1997/04/18  14:58:00  marzetta
 * variable size
 *
 * Revision 1.2  1997/02/26  19:15:35  marzetta
 * defined digits
 *
 * Revision 1.1  1996/06/17  15:50:47  marzetta
 * Initial revision
 *
 * Revision 1.1  1996/06/17  13:39:38  marzetta
 * Initial revision
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "prs_mp.h"

 /* This is the max permitted, the A array is allocated to this size */
long digits = DEFAULT_DIGITS;

/* this is the biggest acheived so far. */
long record_digits = 0;

/**********MACHINE DEPENDENT CONSTANTS***********/
/* MAXD is 2^(k-1)-1 where k=16,32,64 word size */
/* MAXD must be at least 2*BASE^2               */
/* If BASE is 10^k, use "%k.ku" for FORMAT      */
/* INTSIZE is number of bytes for integer       */
/* 16 bit machines                              */
/* remove L from constants, change %ld to %d    */
/*#define MAXD 32767                            */
/*#define BASE 100                              */
/*#define FORMAT "%2.2u"                        */
/*#define INTSIZE 4                             */
/*#define BIT "16"                              */
/* 32/64 bit machines                           */
/* add L to constants, change %d to %ld  */
#ifndef B64
/*32 bit machines */
#define FORMAT "%4.4lu"                    
#define MAXD 2147483647L                      
#define BASE 10000L
#define INTSIZE 8L
const char BIT[] = "32";
#else
/* 64 bit machines */                      
#define MAXD 9223372036854775807L     
#define BASE 1000000000L             
#define FORMAT "%9.9lu"       
#define INTSIZE 16L 
const char BIT[] = "64";
#endif

#define TWO 2L
#define MAXINPUT 1000  /*max length of any input rational*/    


/******* values for GCD *******/
static long MAXSPLEN;		/* Maximum digits for a number that will fit */
				/* into a single long integer. */
static unsigned long MAXSPVAL;	/* Max value for the last digit to guarantee */
				/* fitting into a single long integer. */
static mp mpone;

/***************************************************************/
/*                                                             */
/*     Package of routines for multiple precision arithmetic   */
/*                                                             */
/***************************************************************/
void pmp(const char *name, mp a)  /*print the long precision integer a*/
{
  long i;

  printf("%s",name);
  if (sign(a)==NEG) printf("-");
  printf("%lu",a[length(a)-1]);
  for (i=length(a)-2;i>=1;i--) printf(FORMAT,a[i]);
} /* pmp */

void notimpl(const char *s)
{fprintf(stderr,"\n%s\n",s);
 assert(0);
 exit(1);
}

void copy(mp a, mp b)   /* assigns a=b  */
{
  long i;
  for (i=0;i<=length(b);i++) a[i]=b[i];
}

void storelength(mp a, long la)   /*store length la in a[0], keeping sign*/
{
  if (la > record_digits) {
    record_digits = la;
    if (la > digits)
      { pmp("a=",a);
      printf(" BASE: %ld     DIGITS:  %ld \n",BASE,digits);
      printf(" Maximum integer is BASE^(DIGITS-1)-1 \n");
      notimpl("Overflow!  Increase BASE and/or DIGITS and recompile!");
      }
  }
  if (a[0] < 0 ) a[0]=(-la);
	else a[0] = la;
}

void itomp(long in, mp a)     /* convert integer i to multiple precision */
{
  long i;
  a[0]=2;          /* initialize to zero */
  for(i=1;i<digits;i++) a[i]=0;
  if (in < 0 ) {
    storesign(a,NEG);
    in=in*(-1);
  }
  i=0;
  while (in != 0) {
    i++;
    a[i]= in-BASE*(in/BASE);
    in=in/BASE;
    storelength(a,i+1);
  }
}        /* end of itomp  */

void normalize(mp a)
{
  long cy;
  int i, la;

  la=length(a);
  cy=0;
  for(i=1;i<la;i++) {		/* a[i] + cy */
    cy=(a[i]+=cy)/BASE;
    a[i]-=cy*BASE;
    if(a[i]<0) {
      a[i]+=BASE;cy--;
    }
  }
  /* a[la] + cy */
  if (cy<0) {
    storesign(a,sign(a)==POS ? NEG : POS );

    for (i=1; !a[i] && i<la; i++);
    if (i<la) {
      a[i] = BASE - a[i];
      i++;
      for(;i<la;i++)
	a[i] = BASE - 1 - a[i];
      cy = - cy - 1;
    } else
      cy = - cy;
  }
  while(cy>0) {
    a[i++]=cy%BASE;
    cy/=BASE;
  }
  while (a[i-1]==0 && i>2) i--;
  storelength(a,i);
  if(i==2 && a[1]==0) storesign(a,POS);
}  /* end of normalize */

/********************************************************/
/* Divide two multiple precision integers (c=a/b).      */
/* a is destroyed and contains the remainder on return. */
/* From Knuth Vol.2 SemiNumerical Algorithms            */
/********************************************************/
void divint(mp a , mp b , mp c)
{
  long cy, la, lb, lc, d1, s, t, sig;
  long i, j, qh;
  
/*  figure out and save sign, do everything with positive numbers*/
  sig=sign(a)*sign(b);

  la = length(a);
  lb = length(b);
  lc = la-lb+2;
  if ( la<lb )
  {
    storelength(c,TWO);
    storesign(c,POS);
    c[1] = 0;
    normalize(c);
    return;
  }
  for( i=1; i<lc; i++ ) c[i] = 0;
  storelength(c,lc);
  storesign(c, (sign(a)==sign(b)) ? POS : NEG );

  /******************************/
  /* division by a single word: */
  /*  do it directly            */
  /******************************/

  if( lb==2 ) {
    cy = 0;
    t = b[1];
    for (i=la-1; i>0; i--)  {
      cy = cy*BASE+a[i];
      a[i] = 0;
      cy -= (c[i] = cy/t) * t;
    }
    a[1] = cy;
    storesign(a,(cy==0) ? POS : sign(a));
    storelength(a,TWO);
    /*      set sign of c to sig  (**mod**)            */
    storesign(c,sig);
    normalize(c);
    return;
  }
  else
  {
    /* mp's are actually DIGITS+1 in length, so if length of a or b = */
    /* DIGITS, there will still be room after normalization. */
    /****************************************************/
    /* Step D1 - normalize numbers so b > floor(BASE/2) */
    d1 = BASE/(b[lb-1] + 1);
    if (d1 > 1)
    {
      cy = 0;
      for (i=1;i<la;i++)
      {
	cy = (a[i]=a[i]*d1+cy)/BASE;
	a[i] %= BASE;
      }
      a[i] = cy;
      cy = 0;
      for (i=1;i<lb;i++)
      {
	cy = (b[i]=b[i]*d1+cy)/BASE;
	b[i] %= BASE;
      }
      b[i] = cy;
    }
    else
    {
      a[la] = 0;		/* if la or lb = DIGITS this won't work */
      b[lb] = 0;
    }
    /*********************************************/
    /* Steps D2 & D7 - start and end of the loop */
    for (j = 0;j<=la-lb;j++)
    {
      /*************************************/
      /* Step D3 - determine trial divisor */
      if (a[la-j] == b[lb-1])
	qh = BASE - 1;
      else
      {
	s = (a[la-j]*BASE + a[la-j-1]);
	qh = s/b[lb-1];
	while (qh*b[lb-2] > (s - qh*b[lb-1])*BASE + a[la-j-2])
	  qh--;
      }
      /*******************************************************/
      /* Step D4 - divide through using qh as quotient digit */
      cy = 0;
      for (i=1;i<=lb;i++)
      {
	s = qh*b[i] + cy;
	a[la-j-lb+i] -= s%BASE;
	cy = s/BASE;
	if (a[la-j-lb+i] < 0)
	{
	  a[la-j-lb+i] += BASE;
	  cy++;
	}
      }
      /*****************************************************/
      /* Step D6 - adjust previous step if qh is 1 too big */
      if (cy)
      {
	qh--;
	cy = 0;
	for (i=1;i<=lb;i++)	/* add a back in */
	{
	  a[la-j-lb+i] += b[i] + cy;
	  cy = a[la-j-lb+i]/BASE;
	  a[la-j-lb+i] %= BASE;
	}
      }
      /***********************************************************************/
      /* Step D5 - write final value of qh.  Saves calculating array indices */
      /* to do it here instead of before D6 */
      
      c[la-lb-j+1] = qh;
      
    }
    /**********************************************************************/
    /* Step D8 - unnormalize a and b to get correct remainder and divisor */
    
    for (i=lc;c[i-1]==0 && i>2;i--); /* strip excess 0's from quotient */
    storelength(c,i);
    if(i==2 && c[1]==0) storesign(c,POS);
    cy = 0;
    for (i=lb-1;i>=1;i--)
    {
      cy = (a[i]+=cy*BASE)%d1;
      a[i] /= d1;
    }
    for (i=la;a[i-1]==0 && i>2;i--); /* strip excess 0's from quotient */
    storelength(a,i);
    if(i==2 && a[1]==0) storesign(a,POS);
    if (cy) printf("divide error");
    for (i=lb-1;i>=1;i--)
    {
      cy = (b[i]+=cy*BASE)%d1;
      b[i] /= d1;
    }
  }
} /* divint */

int greater(mp a , mp b)  /* tests if a > b and returns (TRUE=POS) */
{
  long i;

  if(a[0] > b[0]) return(TRUE);
  if(a[0] < b[0]) return(FALSE);

  for (i=length(a)-1;i>=1;i--)
     {
      if(a[i] < b[i]) 
              {
              if(sign(a) == POS) return (0);
                        else return (1);
              }
      if(a[i] > b[i]) 
              {
              if(sign(a) == NEG) return (0);
                        else return (1);
              }
     }
  return(0);
}

void gcd(mp u , mp v)   /*returns u=gcd(u,v) using classic Euclid's algorithm
             v is destoyed.  Knuth, II, p.320  
             modified to avoid copies r=u,u=v,v=r
	     Switches to single precision when possible for greater speed */
{
  mp r;
  unsigned long ul,vl;
  long i;
    
  if(greater(v,u)) goto bigv;
  bigu:
       if(zero(v))  return;
  if ((i=length(u))<MAXSPLEN || (i==MAXSPLEN && u[MAXSPLEN-1]<MAXSPVAL))
    goto quickfinish;
       divint(u,v,r);
       normalize(u);

  bigv:
       if(zero(u)) { copy(u,v); return;}
  if ((i=length(v))<MAXSPLEN || (i==MAXSPLEN && v[MAXSPLEN-1]<MAXSPVAL))
    goto quickfinish;
       divint(v,u,r);
       normalize(v);
 goto bigu;
  /* Base 10000 only at the moment */
  /* when u and v are small enough, transfer to single precision integers */
  /* and finish with euclid's algorithm, then transfer back to mp */
 quickfinish:
  ul = vl = 0;
  for (i=length(u)-1;i>0;i--)
    ul = BASE*ul+u[i];
  for (i=length(v)-1;i>0;i--)
    vl = BASE*vl+v[i];
  if (ul>vl) goto qv;
 qu:
  if (!vl) 
  {
    for (i=1;ul;i++)
    {
      u[i]=ul%BASE;
      ul=ul/BASE;
    }
    storelength(u,i);
    return;
  }
  ul %= vl;
 qv:
  if (!ul) 
  {
    for (i=1;vl;i++)
    {
      u[i]=vl%BASE;
      vl=vl/BASE;
    }
    storelength(u,i);
    return;
  }
  vl %= ul;
  goto qu;
}

void mulint(mp a, mp b, mp c)
     /*** multiply two integers a*b --> c
       Handbook of Algorithms and Data Structures, p239  ***/
{
  long i,j,la,lb, bi;
  /*** b and c may coincide ***/
  la=length(a); lb=length(b);
  if (la+lb-2 > digits ) { 
    printf("\n Mulint overflow for a*b \n");
    pmp("a=",a);
    pmp("b=",b);
    printf(" BASE: %ld     DIGITS:  %ld \n",BASE,digits);
    printf(" Maximum integer is BASE^(DIGITS-1)-1 \n");
    notimpl("Increase BASE and/or DIGITS and recompile!");
  }
  for (i=0;i<la-2;i++) c[lb+i]=0;
  for (i=lb-1;i>0;i--) {
    bi = b[i];
    for(j=2;j<la;j++)
      if((c[i+j-1]+=bi*a[j]) >
         MAXD-(BASE-1)*(BASE-1)-MAXD/BASE){
        c[i+j-1] -= (MAXD/BASE)*BASE;
        c[i+j] += MAXD/BASE;
      }
    c[i] = bi*a[1];
  }
  storelength(c,la+lb-2);
  storesign(c,sign(a)==sign(b) ? POS : NEG );
  normalize(c);
}  /***end of mulint ***/

void lcm(a,b)      /* a = least common multiple of a, b; b is preserved */
     mp a,b;
{
  mp u,v;
  copy(u,a); copy(v,b);
  gcd(u,v);
  divint(a,u,v);   /* v=a/u   a contains remainder = 0 */
  mulint(v,b,a);
}             /* end of lcm */

void linint(mp a , long ka , mp b , long kb)
     /***compute a*ka+b*kb --> a ***/
     /***Handbook of Algorithms and Data Structures P.239 ***/
{
  int i,la,lb;
  
  la=length(a);
  lb=length(b);
  if (sign(a) != sign(b)) kb = -kb;
  if (lb>la) {
    storelength(a,lb);
    for (i=1; i<la; i++) a[i] = ka*a[i] + kb*b[i];
    for (;    i<lb; i++) a[i] = kb*b[i];
  } else {
    for (i=1; i<lb; i++) a[i] = ka*a[i] + kb*b[i];
    for (;    i<la; i++) a[i] *= ka;
  }
  normalize(a);
}   /***end of linint***/

void subint(mp a, mp b)
{
  /* Compute a-b --> a */
  int i,la,lb;
  
  la=length(a);
  lb=length(b);
  if (sign(a) != sign(b)) {	/* addition */
    if (lb>la) {
      storelength(a,lb);
      for (i=1; i<la; i++) a[i] += b[i];
      for (;    i<lb; i++) a[i] = b[i];
    } else {
      for (i=1; i<lb; i++) a[i] += b[i];
    }
  } else {			/* subtraction */
    if (lb>la) {
      storelength(a,lb);
      for (i=1; i<la; i++) a[i] -= b[i];
      for (;    i<lb; i++) a[i] = -b[i];
    } else {
      for (i=1; i<lb; i++) a[i] -= b[i];
    }
  }
  normalize(a);
} /* subint */

/***************************************************************/
/*                                                             */
/*  End of package for multiple precision arithmetic           */
/*                                                             */
/***************************************************************/
/*
zero(a)
mp(a);
{
return		((((a)[0]==2 || (a)[0]==-2) && (a)[1]==0)?TRUE:FALSE);
}
length(a)
mp a;
{
return	(((a)[0] > 0) ? (a)[0] : -(a)[0]) ;
}
sign(a)
mp a;
{
return (((a)[0] < 0) ? NEG : POS);
}
positive(a)
mp a;
{
return	(((a)[0] < 2 || ((a)[0]==2 && (a)[1]==0))?FALSE:TRUE);
}
storesign(a,sa)
mp a;
long sa;
{
return	a[0]=((a)[0] > 0) ? (sa)*((a)[0]) : -(sa)*((a)[0]);
}
negative(a)
mp a;
{
return	(((a)[0] > -2 || ( (a) [0]==-2 &&  (a) [1]==0))?FALSE:TRUE);
}
*/

/***************************************************************/
/*                                                             */
/*     Package of routines for rational arithmetic             */
/*     (Built on top of package for multiprecision arithmetic  */
/*                                                             */
/***************************************************************/

void reduce(mp Na, mp Da)    /* reduces Na/Da by gcd(Na,Da) */
{
  mp Nb,Db,Nc,Dc;
  copy(Nb,Na);  copy(Db,Da); storesign(Nb,POS); storesign(Db, POS);
  copy(Nc,Na);  copy(Dc,Da); 
  gcd(Nb,Db); /* Nb is the gcd(Na,Da) */
  divint(Nc,Nb,Na); divint(Dc,Nb,Da);
}
int comprod(mp Na, mp Nb, mp Nc, mp Nd)      /* +1 if Na*Nb > Nc*Nd  */
                                            /* -1 if Na*Nb < Nc*Nd  */
                                            /*  0 if Na*Nb = Nc*Nd  */
{
  mp mc,md;
  mulint(Na,Nb,mc);
  mulint(Nc,Nd,md);
  subint(mc, md);
  if(positive(mc)) return (1);
  if(negative(mc)) return (-1);
  return(0);
}

void linrat(Na,Da,ka,Nb,Db,kb,Nc,Dc) /* computes Nc/Dc = ka*Na/Da  +kb* Nb/Db
	   			        and reduces answer by gcd(Nc,Dc) */
     mp Na,Da,Nb,Db,Nc,Dc;
     long ka,kb;
{
  mp c;
  mulint(Na,Db,Nc);
  mulint(Da,Nb,c);
  linint(Nc,ka,c,kb);  /* Nc = (ka*Na*Db)+(kb*Da*Nb)  */
  mulint(Da,Db,Dc);  /* Dc =  Da*Db           */
  reduce(Nc,Dc);
}


void divrat(Na,Da,Nb,Db,Nc,Dc) /* computes Nc/Dc = (Na/Da)  / ( Nb/Db )
                                  and reduces answer by gcd(Nc,Dc) */
     mp Na,Da,Nb,Db,Nc,Dc;
{
  mulint(Na,Db,Nc);
  mulint(Da,Nb,Dc);
  reduce(Nc,Dc);
}
 

void mulrat(Na,Da,Nb,Db,Nc,Dc) /* computes Nc/Dc = Na/Da  * Nb/Db
                                  and reduces answer by gcd(Nc,Dc) */
     mp Na,Da,Nb,Db,Nc,Dc;
{
  mulint(Na,Nb,Nc);
  mulint(Da,Db,Dc);
  reduce(Nc,Dc);
}

/***************************************************************/
/*                                                             */
/*     End package of routines for rational arithmetic         */
/*                                                             */
/***************************************************************/

void atomp(const char *s, mp a)  /*convert string to mp integer*/
     /* based on  atoi KR p.58 */
{
  long diff,ten,i,sig;
  ten=10L;
  for (i=0; s[i]==' ' || s[i]=='\n' || s[i]=='\t'; i++);
  /*skip white space*/
  sig=POS;
  if( s[i] == '+' || s[i]=='-' ) /* sign */
    sig=(s[i++]=='+') ? POS : NEG;
  itomp(0L,a);
  while ( s[i] >= '0' && s[i] <= '9')
    { diff=s[i] - '0';
    linint(a,ten,mpone,diff);
    i++;
    }
  storesign(a,sig);
  if (s[i]) {
    fprintf(stderr, "\nIllegal character in number: '%s'\n", s+i);
    exit(1);
  }
}  /* end of atomp */

void atoaa(const char *in, char *num, char *den)
     /* convert rational string in to num/den strings*/
{
 long i,j;
 for(i=0;in[i] != '\0' && in[i] != '/';i++) num[i]=in[i];
 num[i]='\0';
 den[0]='\0';
 if (in[i] == '/')
    {
     for(j=0;in[j+i+1] != '\0' ;j++) den[j]=in[i+j+1];
     den[j]='\0';
    }
}               /* end of atoaa */

int readrat(mp Na, mp Da) /* read a rational or integer and convert to mp */
               /* returns true if denominator is not one       */
{
  char in[MAXINPUT],num[MAXINPUT],den[MAXINPUT];
  scanf("%s",in);
  atoaa(in,num,den);     /*convert rational to num/dem strings*/
  atomp(num,Na);
  if (den[0]=='\0') 
        {
        itomp(1L,Da);
        return(FALSE);
        }
   atomp(den,Da);
   return(TRUE);
}

void readmp(mp a)      /* read an integer and convert to mp */
{
  long in;
  scanf("%ld",&in);
  itomp(in,a);
}

void prat(const char *name, mp Nt, mp Dt)
     /*print the long precision rational Nt/Dt  */
{
  pmp(name, Nt);
  if( !(Dt[0]==2 && Dt[1]==1)) pmp("/", Dt);
  printf(" ");
} /* prat */

void prs_mp_Init(void)
{
  itomp(1,mpone);
  MAXSPVAL = MAXD;
  for (MAXSPLEN = 2;MAXSPVAL >= BASE; MAXSPLEN++)
    MAXSPVAL /= BASE;
} /* prs_mp_Init */
