/* prs_mp.h (multiprecision and rational arithmetic)
   Ambros Marzetta
   derived from David Avis' lrs.c, version 2.5i, 1995-08-15
  
 * $Source: /home/schilte/marzetta/work/pd/RCS/prs_mp.h,v $
 * $Revision: 1.4 $
 * $Date: 1997/10/01 16:54:35 $
 * $Log: prs_mp.h,v $
 * Revision 1.4  1997/10/01  16:54:35  marzetta
 * *** empty log message ***
 *
 * Revision 1.3  1997/04/18  14:58:00  marzetta
 * variable size
 *
 * Revision 1.2  1997/02/26  19:15:35  marzetta
 * defined digits
 *
 * Revision 1.1  1996/06/17  15:50:47  marzetta
 * Initial revision
 *
 */

/* Type ----------------------------------------------------------------- */

/*
   this is in decimal digits, you pay in memory if you increase this,
   unless you override by a line with 
   digits n
   before the begin line of your file.
*/
#define DEFAULT_DIGITS 180L

/* 
   this is number of longwords. Increasing this won't cost you that much
   since only variables other than the A matrix are allocated this size.
   Changing affects running time in small but not very predictable ways.
*/

#define MAX_DIGITS 255L 

/* 
 *  convert between decimal and machine (longword digits). Notice lovely
 *  implementation of ceiling function :-)
 */

#define DEC2DIG(d) ( (d) % BASE_DIG ? (d)/BASE_DIG+1 : (d)/BASE_DIG)
#define DIG2DEC(d) ((d)*BASE_DIG)

typedef long mp[MAX_DIGITS+1]; /* type mp holds one multi-precision integer */

 /* This is the max permitted, the A array is allocated to this size */
extern long digits;

/* this is the biggest acheived so far. */
extern long record_digits;  

#ifdef __alpha
#define B64
#endif

#ifndef B64
/*32 bit machines */
#define BASE_DIG 4   
#else
/* 64 bit machines */                      
#define BASE_DIG 9
#endif

/* Macros --------------------------------------------------------------- */

#ifndef TRUE
#define TRUE 1L
#endif
#ifndef FALSE
#define FALSE 0L
#endif

#define POS 1L
#define NEG -1L

#define positive(a)     (((a)[0] < 2 || ((a)[0]==2 && (a)[1]==0))?FALSE:TRUE)
#define negative(a)     (((a)[0] > -2 || ((a)[0]==-2 && (a)[1]==0))?FALSE:TRUE)
#define length(a)       (((a)[0] > 0) ? (a)[0] : -(a)[0])
#define sign(a)         (((a)[0] < 0) ? NEG : POS)
#define storesign(a,sa) a[0]=((a)[0] > 0) ? (sa)*((a)[0]) : -(sa)*((a)[0])
#define changesign(a)   a[0]= -(a)[0]
#define zero(a)         ((((a)[0]==2 || (a)[0]==-2) && (a)[1]==0)?TRUE:FALSE)

extern const char BIT[];


/* Functions ------------------------------------------------------------ */

void pmp(const char * name , mp a);
void notimpl(const char * s);
void copy(mp a , mp b);
void storelength(mp a , long la);
void itomp(long in, mp a);
void normalize(mp a);
void divint(mp a , mp b , mp c);
int greater(mp a , mp b);
void gcd(mp u , mp v);
void mulint(mp a , mp b , mp c);
void lcm(mp a , mp b);
void linint(mp a , long ka , mp b , long kb);
void subint(mp a , mp b);
void reduce(mp Na , mp Da);
int comprod(mp Na , mp Nb , mp Nc , mp Nd);
void linrat(mp Na, mp Da, long ka, mp Nb, mp Db, long kb, mp Nc, mp Dc);
void divrat(mp Na , mp Da , mp Nb , mp Db , mp Nc , mp Dc);
void mulrat(mp Na , mp Da , mp Nb , mp Db , mp Nc , mp Dc);
void prat(const char * name , mp Nt , mp Dt);
int readrat(mp Na , mp Da);

/* Initialization -------------------------------------------------------- */

void prs_mp_Init(void);
